import numpy as np
import math
import tqdm
import random

pi = np.pi
e = np.e
sqrt = np.sqrt
log = np.log2
ln = lambda x: np.log(float(x))
ceil = np.ceil
floor = np.floor

def convert_size(size_bits):
    if size_bits == 0:
        return "0B"
    size_bytes = size_bits / 8
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])


class Parameters:
    def __init__(self, secparam, interactive, logQ, n, m, N, d, logq, logbase, l, k, opt_mode='Aggressive', reduction_mode='Proven', domain='Zq') -> None:

        if opt_mode not in ['Aggressive', 'Conservative']:
            raise ValueError("Invalid opt_mode: {}".format(opt_mode))
        if opt_mode == 'Aggressive':
            self.aggressive_sis = True
        else:
            self.aggressive_sis = False
        self.opt_mode = opt_mode

        if domain not in ['Rq', 'Zq']:
            raise ValueError("Invalid domain: {}".format(domain))
        if domain == 'Rq':
            self.rq = True
        else:
            assert((d+1) % N == 0)
            self.rq = False
        self.domain = domain

        if reduction_mode not in ['Proven', 'Heuristic']:
            raise ValueError("Invalid reduction_mode: {}".format(reduction_mode))
        if reduction_mode == 'Proven':
            self.heuristic = False
        else:
            self.heuristic = True
        self.reduction_mode = reduction_mode 

        # Degree of polynomial to commit
        self.orig_d = d
        self.d = d if self.rq else ((d+1) // N) - 1
        self.h = round(log(self.d+1))
        # We want things to split nicely
        assert(self.d >= 1)
        assert(2**self.h == self.d+1)
        # Security parameter and number of RO queries
        self.secparam = secparam
        self.interactive = interactive
        if not self.interactive:
            self.logQ = logQ

        # Dimension of SIS matrices
        self.n = n
        self.m = m

        # Ring dimensions
        self.logq = logq
        self.N = N

        # Dimensions to use for the base
        self.logbase = logbase

        # Folding factor
        self.k = k
        # We want things to split nicely
        assert(self.h % self.k == 0)
        # Rounds to apply
        self.l = min(l, self.h) # Don't recurse too much

        # Dynamically computed values start here
        self.logd = log(self.d+1)
        self.logn = log(self.n)
        self.logm = log(self.m)
        self.logN = log(self.N)
        self.kpow = 2**self.k

        # Used for MP12 initial size
        self.qp = self.logq / self.logbase
        self.t = self.n * self.qp
        assert(self.t + self.n <= self.m)
        self.mp = 2 * self.m + self.t
        self.np = 2 * self.t
        self.tp = max(self.np, self.mp)
        
        # Compute the extraction standard deviation
        self.logmsis_sigma_0 = self.msis_sigma_0()

        self.standard = self.stddev()

        self.r = self.repetitions()
        self.logr = log(self.r)

        self.logbeta_starting = self.beta_starting()
        self.logbeta_completeness = self.beta_completeness()
        self.loggamma_extraction = self.gamma_extraction()
        self.loghprisis_bound = self.hprisis_bound()
        self.msis_red_m = self.msis_m()

    def repetitions(self) -> int:
        r = 1
        while True:
            if self.bit_security(r) > self.secparam:
                return r
            else:
                r += 1

    def bit_security(self, r) -> float:
        # Lemma 5.8 + Lemma 2.28
        logr = log(r)
        if self.interactive:
            # In this case, knowledge error is l r 2^k / (2N)^r
            return r * (1 + self.logN) - (log(self.l) + logr + self.k)
        else:
            # In this case, knowledge error is (Q + 1) l r 2^k / (2N)^r
            return r * (1 + self.logN) - (self.logQ + log(self.l) + logr + self.k)

    # Norms are all in log 
    def stddev(self) -> float:
        # Lemma 4.1
        return 1 + self.logN + log(self.N) + self.logq * (self.n/(self.m-self.t) + 2/(self.N*(self.m - self.t)))

    def beta_starting(self) -> float:
        # Lemma 4.1
        logt = log(self.t)
        logmp = log(self.mp)
        lognp = log(self.np)
        sigma_0 = 1 + self.logbase + self.standard + self.logN + (0.5) * (logt + log(self.m - self.t) + log(log(self.tp * self.N)))
        # If proven we need to increase sigma_0
        if not self.heuristic:
            sigma_0 = max(sigma_0, self.logmsis_sigma_0) # Check deps

        # Deterministic sampling
        logqp = self.logq / self.logbase
        beta_s = sigma_0 + (0.5) * (logmp + lognp + self.logN + 1 + self.logn + logqp) + self.logbase + self.logN
        # Probabilistic sampling
        #sigma_1 = self.logbase + sigma_0 + self.logN + (0.5) * (logmp + lognp + log(log(self.tp * self.N)))
        #beta_s = sigma_1 + (0.5) * (logmp + self.logN)
        
        return beta_s

    def beta_completeness(self) -> float:
        # Lemma 5.4
        return self.l * (self.logr + self.k) + self.logbeta_starting

    def gamma_extraction(self) -> float:
        # Lemma 5.8
        return self.l * ((self.k+1) + self.logr + self.logN) + self.logbeta_starting

    def hprisis_bound(self) -> float:
        # Lemma 5.8
        return 1 + self.loggamma_extraction + sqrt(self.h)

    # Now, we compute the norms for the reduction for MSIS
    def msis_secparam(self):
        # Lemma 3.3 (the guessing part)
        return self.secparam + self.logd

    def msis_m(self):
        # Lemma 3.3
        return self.m if self.heuristic else self.m * self.h

    def msis_sigma_0(self):
        # Lemma 3.3
        logt = log(self.t)
        epsilon = (self.msis_secparam() + 10 * self.N) / (self.N * self.logq)
        epsilon = min(epsilon, 1/3)
        tau = max(2 * self.n, 2 * self.m + self.t)
        l = (0.5) * (self.logN + log(ln(8) + ln(2) * self.logN + ln(2)*self.logq)) + (0.5 + epsilon) * self.logq # With sqrt optimization
        r = (1.5) * (self.logN + log(self.logN * ln(2)))
        msis_stdev = max(l, r)
        # Ignoring the + 2 * t in the log as it should be small, and otherwise easy to overflow
        return self.logbase + (0.5) * (logt + self.logN + 2*self.logN + 2*msis_stdev + self.logm) + (0.5) * (self.logN + log(log(tau * self.N)))


    # Function for estimating the MSIS hardness given parameters:
    # a (n x m) matrix in \Rq along with the solution bound B. It returns the
    # root Hermite factor \delta. We use the methodology presented by
    # [GamNgu08, Mic08] and hence m is irrelevant.
    def findMSISdelta(self):
        logC = self.loghprisis_bound
        logdelta = logC**2 / (4*self.n*self.N*self.logq)
        return logdelta

    #root Hermite factor
    def RHF(self, block):
        return ((pi*block)**(1/block) * block/(2*pi*e))**(1/(2*(block-1)))

    #BKZ block size
    def blocksize(self):
        if self.aggressive_sis:
            #naively looking for a block
            for block in range(200,1000):
                runtime = (8*self.msis_red_m*self.d) * 2**(0.292 * block + 16.4)
                if log(runtime) >= self.msis_secparam():
                    return block
        else:
            return self.msis_secparam()/0.292

    #B is bound, n is the height of the SIS matrix, d is the ring dimension and q is the modulus
    def isMSISSecure(self):
        block = self.blocksize()
        rhf = self.RHF(block)
        if self.findMSISdelta() < log(rhf) and self.loghprisis_bound < self.logq:   
            return True 
        else:
            return False

    def prover_communication_complexity(self) -> int:
        # Lemma 5.9
        size_of_Rq = self.logq * self.N
        eval_points = (self.l + 1) * self.kpow * size_of_Rq
        short_openings = 2 * self.kpow * self.m * self.N * sum(ceil(1 + i * (self.logr + self.k) + self.logbeta_starting) for i in range(self.l))
        final_message = 2**(self.h - self.k * self.l) * size_of_Rq + 2**(self.h - self.k * self.l + 1) * ceil(1 + self.l * (self.logr + self.k) + self.logbeta_starting)
        return eval_points + short_openings + final_message

    def verifier_communication_complexity(self) -> int:
        return self.l * (self.r**2) * self.kpow * ceil(1 + self.logN)
    
    def print(self):
        print(f'l={self.l}, d={self.d}, secparam={self.secparam}, n={self.n}, m={self.m}, N={self.N}, base={self.logbase}, logq={self.logq}, extracted_norm={round(self.loggamma_extraction)}, beta={round(self.logbeta_starting)}, beta_l={round(self.logbeta_completeness)}, repetitions={self.r}, pcc={convert_size(self.prover_communication_complexity())}, vcc={convert_size(self.verifier_communication_complexity())}')

    # Greedly shrink the parameters as much as possible
    def shrink(self):
        assert(self.isMSISSecure())
        params = self
        while True:
            newlogq = params.logq - 1
            test = Parameters(secparam=params.secparam, interactive=params.interactive, logQ=params.logQ, N=params.N, d=params.orig_d, n=params.n, m=params.m, logq=newlogq, logbase=params.logbase, l=params.l, k=params.k, opt_mode=params.opt_mode, reduction_mode=params.reduction_mode, domain=params.domain)
            if test.isMSISSecure():
                params = test
            else:
                break

        while True:
            newn = params.n - 1
            test = Parameters(secparam=params.secparam, interactive=params.interactive, logQ=params.logQ, N=params.N, d=params.orig_d, n=newn, m=params.m, logq=params.logq, logbase=params.logbase, l=params.l, k=params.k, opt_mode=params.opt_mode, reduction_mode=params.reduction_mode, domain=params.domain)
            if test.isMSISSecure():
                params = test
            else:
                break

        while True:
            newm = params.m - 1
            try:
                test = Parameters(secparam=params.secparam, interactive=params.interactive, logQ=params.logQ, N=params.N, d=params.orig_d, n=params.n, m=newm, logq=params.logq, logbase=params.logbase, l=params.l, k=params.k, opt_mode=params.opt_mode, reduction_mode=params.reduction_mode, domain=params.domain)
            except AssertionError:
                break 
            if test.isMSISSecure():
                params = test
            else:
                break

        return params


# This function does a randomised search to find best parameters
def find_parameters_random(secparam, interactive, logQ, d, k, l, opt_mode, reduction_mode, iters=2**18):
    valid = []
    for _ in tqdm.tqdm(range(iters)):
        logq = random.choice(range(200, 1024))
        n = random.choice(range(1, 2000))
        gamma = random.uniform(1, 20)
        logbase = random.choice(range(1, round(logq)))
        qp = floor(logq / logbase)
        t = n * qp
        m = ceil(t + gamma * n) # + log(secparam))
        logN = random.choice(range(1, 30))

        try:
            params = Parameters(secparam=secparam, interactive=interactive, logQ=logQ, N=2**logN, d=d, n=n, m=m, logq=logq, logbase=logbase, l=l, k=k, opt_mode=opt_mode, reduction_mode=reduction_mode)
            if params.isMSISSecure():
                valid.append(params)
        except AssertionError:
            pass

    valid.sort(key=lambda tup: tup.prover_communication_complexity())
    for i in range(min(5, len(valid))):
        valid[i].print()

    valid = [valid.shrink() for valid in valid[:20]]
    valid.sort(key=lambda tup: tup.prover_communication_complexity())
    print('---')
    for i in range(min(5, len(valid))):
        valid[i].print()
        

# This function does a brute force search to find best parameters
def find_parameters_bruteforce(secparam, interactive, logQ, d, k, l, opt_mode, reduction_mode):
    valid = []
    for logN in tqdm.tqdm([5, 6, 7, 8]):
        for n in range(150, 600, 50):
            for logq in range(250,550,20):
                for logbase in range(2,round(logq),1):
                    qp = ceil(logq / logbase)
                    t = n * qp
            
                    # instead of log(secparam), I will put 0 -- for q > 2^200 this is enough
                    # I will ignore gamma because I couldn't find the reason
                    m = ceil(t + n + 1)    
                    try:
                        params = Parameters(secparam=secparam, interactive=interactive, logQ=logQ, N=2**logN, d=d, n=n, m=m, logq=logq, logbase=logbase, l=l, k=k, opt_mode=opt_mode, reduction_mode=reduction_mode)
                        if params.isMSISSecure():
                            valid.append(params)
                    except AssertionError:
                        pass
    valid.sort(key=lambda tup: tup.prover_communication_complexity())
    for i in range(min(5,len(valid))):
        valid[i].print()

    valid = [valid.shrink() for valid in valid[:40]]
    valid.sort(key=lambda tup: tup.prover_communication_complexity())
    print('---')
    for i in range(min(5,len(valid))):
        valid[i].shrink().print()

if __name__ == '__main__':
    pass
    # Example to reproduce the results from the paper
    #find_parameters_random(secparam=128, interactive=False, logQ=64, d=2**20-1, k=1, l=1, opt_mode='Aggressive', reduction_mode='Proven', iters=2**16)
    #find_parameters_bruteforce(secparam=128, interactive=False, logQ=64, d=2**20-1, k=1, l=1, opt_mode='Aggressive', reduction_mode='Proven')
    #find_parameters_random(secparam=128, interactive=False, logQ=64, d=2**30-1, k=1, l=8, opt_mode='Aggressive', reduction_mode='Proven', iters=2**16)
    #find_parameters_bruteforce(secparam=128, interactive=False, logQ=64, d=2**30-1, k=1, l=8, opt_mode='Aggressive', reduction_mode='Proven')
    # Note, the 2^30 example gives slightly worse sizes than in the paper. I have not optimised the bruteforces for this setting, focusing on 2^20. 
